""" XGM related sub-routines

    Copyright (2019) SCS Team.

    (contributions preferrably comply with pep8 code structure
    guidelines.)
"""

import logging

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt

from ..misc.bunch_pattern_external import is_sase_1, is_sase_3
from ..mnemonics_machinery import (mnemonics_to_process,
                                   mnemonics_for_run)


log = logging.getLogger(__name__)


def get_xgm(run, mnemonics=None, merge_with=None, keepAllSase=False,
            indices=slice(0, None)):
    """
    Load and/or computes XGM data. Sources can be loaded on the
    fly via the key argument, or processed from an existing data set
    (merge_with). The bunch pattern table is used to assign the pulse
    id coordinates.

    Parameters
    ----------
    run: extra_data.DataCollection
        DataCollection containing the digitizer data.
    mnemonics: str or list of str
        mnemonics for XGM, e.g. "SCS_SA3" or ["XTD10_XGM", "SCS_XGM"].
        If None, defaults to "SCS_SA3" in case no merge_with dataset
        is provided.
    merge_with: xarray Dataset
        If provided, the resulting Dataset will be merged with this
        one. The XGM variables of merge_with (if any) will also be
        computed and merged.
    keepAllSase: bool
        Only relevant in case of sase-dedicated trains. If True, all
        trains are kept, else only those of the bunchPattern are kept.
    indices: slice, list, 1D array
        Pulse indices of the XGM array in case bunch pattern is missing.
    Returns
    -------
    xarray Dataset with pulse-resolved XGM variables aligned,
     merged with Dataset *merge_with* if provided.

    Example
    -------
    >>> import toolbox_scs as tb
    >>> import toolbox_scs.detectors as tbdet
    >>> run, _ = tb.load(2212, 213)
    >>> xgm = tbdet.get_xgm(run)

    """
    # get the list of mnemonics to process
    mnemonics = mnemonics_to_process(mnemonics, merge_with, 'XGM')

    if len(mnemonics) == 0:
        log.info('No array with unaligned XGM peaks to extract. Skipping.')
        return merge_with
    else:
        log.info(f'Extracting XGM data from {mnemonics}.')

    # Prepare the dataset of non-XGM data to merge with
    if bool(merge_with):
        ds = merge_with.drop(mnemonics, errors='ignore')
    else:
        ds = xr.Dataset()

    run_mnemonics = mnemonics_for_run(run)
    # check if bunch pattern table exists
    if bool(merge_with) and 'bunchPatternTable' in merge_with:
        bpt = merge_with['bunchPatternTable']
        log.debug('Using bpt from merge_with dataset.')
    elif 'bunchPatternTable' in run_mnemonics:
        bpt = run.get_array(*run_mnemonics['bunchPatternTable'].values())
        log.debug('Loaded bpt from DataCollection.')
    else:
        bpt = None

    # Load the arrays, assign pulse ID and merge
    for m in mnemonics:
        if bool(merge_with) and m in merge_with:
            arr = merge_with[m]
            log.debug(f'Using {m} from merge_with dataset.')
        else:
            arr = run.get_array(*run_mnemonics[m].values(), name=m)
            log.debug(f'Loading {m} from DataCollection.')
        if bpt is not None:
            arr = align_xgm_array(arr, bpt)
        else:
            arr = arr.where(arr != 1., drop=True).sel(XGMbunchId=indices)
        ds = ds.merge(arr, join='inner')
    return ds


def align_xgm_array(xgm_arr, bpt):
    """
    Assigns pulse ID coordinates to a pulse-resolved XGM array, according to
    the bunch pattern table. If the arrays contains both SASE 1 and SASE 3
    data, it is split in two arrays.

    Parameters
    ----------
    xgm_arr: xarray DataArray
        array containing pulse-resolved XGM data, with dims ['trainId',
        'XGMbunchId']
    bpt: xarray DataArray
        bunch pattern table

    Returns
    -------
    xarray Dataset with pulse ID coordinates. For SASE 1 data, the coordinates
    name is sa1_pId, for SASE 3 data, the coordinates name is sa3_pId.
    """
    key = xgm_arr.name
    compute_sa1 = False
    compute_sa3 = False
    valid_tid = np.intersect1d(xgm_arr.trainId, bpt.trainId,
                               assume_unique=True)
    # get the relevant masks for SASE 1 and/or SASE3
    if "SA1" in key or "SA3" in key:
        if "SA1" in key:
            mask = is_sase_1(bpt.sel(trainId=valid_tid))
            compute_sa1 = True
        else:
            mask = is_sase_3(bpt.sel(trainId=valid_tid))
            compute_sa3 = True
        tid = mask.where(mask.sum(dim='pulse_slot') > 0, drop=True).trainId
        mask = mask.sel(trainId=tid)
        mask_sa1 = mask.rename({'pulse_slot': 'sa1_pId'})
        mask_sa3 = mask.rename({'pulse_slot': 'sa3_pId'})
    if "XGM" in key:
        compute_sa1 = True
        compute_sa3 = True
        mask_sa1 = is_sase_1(bpt.sel(trainId=valid_tid))
        mask_sa3 = is_sase_3(bpt.sel(trainId=valid_tid))
        mask = xr.ufuncs.logical_or(mask_sa1, mask_sa3)
        tid = mask.where(mask.sum(dim='pulse_slot') > 0,
                         drop=True).trainId
        mask_sa1 = mask_sa1.sel(trainId=tid).rename({'pulse_slot': 'sa1_pId'})
        mask_sa3 = mask_sa3.sel(trainId=tid).rename({'pulse_slot': 'sa3_pId'})
        mask = mask.sel(trainId=tid)

    npulses_max = mask.sum(dim='pulse_slot').max().values
    xgm_arr = xgm_arr.sel(trainId=tid).isel(
                XGMbunchId=slice(0, npulses_max))
    # In rare cases, some xgm data is corrupted: trainId is valid but values
    # are inf / NaN. We set them to -1 to avoid size mismatch between xgm and
    # bpt. Before returning we will drop them.
    xgm_arr = xgm_arr.where(np.isfinite(xgm_arr)).fillna(-1.)
    # pad the xgm array to match the bpt dims, flatten and
    # reorder xgm array to match the indices of the mask
    xgm_flat = np.hstack((xgm_arr.fillna(1.),
                          np.ones((xgm_arr.sizes['trainId'],
                                   2700-npulses_max)))).flatten()
    xgm_flat_arg = np.argwhere(xgm_flat != 1.)
    mask_flat = mask.values.flatten()
    mask_flat_arg = np.argwhere(mask_flat)
    if(xgm_flat_arg.shape != mask_flat_arg.shape):
        log.warning(f'{key}: XGM data and bunch pattern do not match.')
    new_xgm_flat = np.ones(xgm_flat.shape)
    new_xgm_flat[mask_flat_arg] = xgm_flat[xgm_flat_arg]
    new_xgm = new_xgm_flat.reshape((xgm_arr.sizes['trainId'], 2700))
    # create a dataset with new_xgm array masked by SASE 1 or SASE 3
    xgm_dict = {}
    if compute_sa1:
        sa1_xgm = xr.DataArray(new_xgm, dims=['trainId', 'sa1_pId'],
                               coords={'trainId': xgm_arr.trainId,
                                       'sa1_pId': np.arange(2700)},
                               name=key.replace('XGM', 'SA1'))
        sa1_xgm = sa1_xgm.where(mask_sa1, drop=True)
        sa1_xgm = sa1_xgm.where(sa1_xgm != -1., drop=True)
        # remove potential corrupted data:
        xgm_dict[sa1_xgm.name] = sa1_xgm
    if compute_sa3:
        sa3_xgm = xr.DataArray(new_xgm, dims=['trainId', 'sa3_pId'],
                               coords={'trainId': xgm_arr.trainId,
                                       'sa3_pId': np.arange(2700)},
                               name=key.replace('XGM', 'SA3'))
        sa3_xgm = sa3_xgm.where(mask_sa3, drop=True)
        # remove potential corrupted data:
        sa3_xgm = sa3_xgm.where(sa3_xgm != -1., drop=True)
        xgm_dict[sa3_xgm.name] = sa3_xgm
    ds = xr.Dataset(xgm_dict)
    return ds


def calibrate_xgm(run, data, xgm='SCS', plot=False):
    """
    Calculates the calibration factor F between the photon flux (slow signal)
    and the fast signal (pulse-resolved) of the sase 3 pulses. The calibrated
    fast signal is equal to the uncalibrated one multiplied by F.

    Parameters
    ----------
    run: extra_data.DataCollection
        DataCollection containing the digitizer data.
    data: xarray Dataset
        dataset containing the pulse-resolved sase 3 signal, e.g. 'SCS_SA3'
    xgm: str
        one in {'XTD10', 'SCS'}
    plot: bool
        If True, shows a plot of the photon flux, averaged fast signal and
        calibrated fast signal.

    Returns
    -------
    F: float
        calibration factor F defined as:
        calibrated XGM [microJ] = F * fast XGM array ('SCS_SA3' or 'XTD10_SA3')

    Example
    -------
    >>> import toolbox_scs as tb
    >>> import toolbox_scs.detectors as tbdet
    >>> run, data = tb.load(900074, 69, ['SCS_XGM'])
    >>> ds = tbdet.get_xgm(run, merge_with=data)
    >>> F = tbdet.calibrate_xgm(run, ds, plot=True)
    >>> # Add calibrated XGM to the dataset:
    >>> ds['SCS_SA3_uJ'] = F * ds['SCS_SA3']
    """

    run_mnemonics = mnemonics_for_run(run)
    # check if bunch pattern table exists
    if 'bunchPatternTable' in data:
        bpt = data['bunchPatternTable']
    elif 'bunchPatternTable' in run_mnemonics:
        bpt = run.get_array(*run_mnemonics['bunchPatternTable'].values())
    elif 'bunchPatternTable_SA3' in run_mnemonics:
        bpt = run.get_array(*run_mnemonics['bunchPatternTable_SA3'].values())
    else:
        raise ValueError('Bunch pattern missing. Cannot calibrate XGM.')
    mask_sa3 = is_sase_3(bpt.sel(trainId=data.trainId))
    npulses_sa3 = np.unique(mask_sa3.sum(dim='pulse_slot'))
    if len(npulses_sa3) == 1:
        npulses_sa3 = npulses_sa3[0]
    else:
        log.warning('change of pulse pattern in sase3 during the run.')
        npulses_sa3 = max(npulses_sa3)
    mask_sa1 = is_sase_1(bpt.sel(trainId=data.trainId))
    npulses_sa1 = np.unique(mask_sa1.sum(dim='pulse_slot'))
    if len(npulses_sa1) == 1:
        npulses_sa1 = npulses_sa1[0]
    else:
        log.warning('change of pulse pattern in sase1 during the run.')
        npulses_sa1 = max(npulses_sa1)

    pflux_key = f'{xgm}_photonFlux'
    if pflux_key in data:
        pflux = data[pflux_key]
    else:
        pflux = run.get_array(*run_mnemonics[pflux_key].values())
        pflux = pflux.sel(trainId=data.trainId)
    pflux_sa3 = (npulses_sa1 + npulses_sa3) * pflux / npulses_sa3
    avg_fast = data[f'{xgm}_SA3'].rolling(trainId=200).mean().mean(axis=1)
    calib = np.nanmean(pflux_sa3.values / avg_fast.values)
    if plot:
        plot_xgm_calibration(xgm, pflux, pflux_sa3, avg_fast, calib)
    return calib


def plot_xgm_calibration(xgm, pflux, pflux_sa3, avg_fast, calib):
    plt.figure(figsize=(8, 4))
    plt.plot(pflux, label='photon flux all')
    plt.plot(pflux_sa3, label='photon flux SA3')
    plt.plot(avg_fast, label='avg pulsed XGM')
    plt.plot(avg_fast*calib, label='calibrated avg pulsed XGM')
    plt.title(f'calibrated XGM = {xgm}_SA3 * {calib:.3e}')
    plt.xlabel('train number')
    plt.ylabel(r'Pulse energy [$\mu$J]')
    plt.legend()
    return
